# encoding: utf-8

from __future__ import unicode_literals

from abc import ABCMeta
from abc import abstractmethod
from abc import abstractproperty
from collections import OrderedDict
from datetime import date
from datetime import datetime

from six import text_type
from six.moves import zip
from xlrd import open_workbook
from xlrd.xldate import xldate_as_tuple

from ..validation import Validator


class Row(object):
    """Строка данных."""

    row_number = None
    values = None

    def __unicode__(self):
        return 'Строка {0}'.format(self.row_number)

    __str__ = __unicode__


class ExcelRow(Row):
    """Строка данных при импорте из Excel."""

    def __init__(self):
        super(ExcelRow, self).__init__()
        self.sheet_name = None

    def __unicode__(self):
        return 'Лист "{0}", {1}'.format(
            self.sheet_name, super(ExcelRow, self).__unicode__()
        )

    __str__ = __unicode__


class IParser(object):
    """Интерфейс парсера."""

    __metaclass__ = ABCMeta

    row_cls = None
    validator_cls = None

    @abstractproperty
    def extensions(self):
        """Список расширений фалов, поддерживаемых парсером."""

    @abstractmethod
    def get_rows(self):
        """Возвращает коллекцию строк."""


class ExcelParser(IParser):
    """Парсер файлов Excel."""

    row_cls = ExcelRow
    validator_cls = Validator

    extensions = ('xls', 'xlsx')

    def __init__(
        self, file_path, rules, logger, skip_rows=None, skip_sheets=None
    ):
        self.file_path = file_path
        self.rules = rules
        self.logger = logger

        self.header_row = 0  # Номер строки шапки
        # Какие строки пропускаем
        self.skip_rows = tuple(skip_rows or ()) + (self.header_row, )

        self.skip_sheets = skip_sheets or ()
        self._workbook = open_workbook(file_path)
        self._header_config = {rules[key]['header']: key for key in rules}

    def get_sheets(self):
        """Возвращает список листов."""
        for sheet in self._workbook.sheets():
            if sheet.name not in self.skip_sheets:
                yield sheet

    @property
    def headers(self):
        """Имена колонок таблицы."""
        sheet = next(self.get_sheets())
        return [cell.value.strip() for cell in sheet.row(self.header_row)]

    def _get_validator(self, rules):
        validator = self.validator_cls(rules)
        return validator

    def _get_row(self):
        return self.row_cls()

    def _prepare_row(self, row):
        values = getattr(row, 'values', {})

        for attname, value in values.items():
            rules = self.rules[attname]
            if 'type' in rules and not isinstance(value, rules['type']):
                try:
                    if rules['type'] is date:
                        values[attname] = datetime(
                            *xldate_as_tuple(value, self._workbook.datemode)
                        ).date()
                    else:
                        values[attname] = rules['type'](value)
                # Пропускаем ошибку конвертации, она детектируется валидатором
                except Exception:  # pylint: disable=W0703
                    pass

    def get_rows(self):
        """Возвращает коллекцию валидных строк из файла."""
        super(ExcelParser, self).get_rows()

        validator = self._get_validator(self.rules)

        for sheet in self.get_sheets():
            for row_number, data in enumerate(sheet.get_rows(), 1):
                if row_number - 1 in self.skip_rows:
                    continue
                row = self._get_row()

                row.sheet_name = sheet.name
                row.row_number = row_number
                row.values = OrderedDict(zip(
                    (self._header_config[header] for header in self.headers),
                    (cell.value for cell in data)
                ))

                self._prepare_row(row)

                try:
                    validator.validate(row)
                except Exception as exception:  # pylint: disable=W0703
                    # TODO: error as object?
                    self.logger.log(
                        exception, prefix=text_type(row)
                    )
                else:
                    yield row
